"""
List of parsers for file formats
"""

from .castep_file_parser import parse_castep_file
from .cell_param_file_parser import parse_cell_param_file
from .md_geom_file_parser import parse_md_geom_file
from .extra_files_parser import (parse_bands_file, parse_hug_file, parse_phonon_dos_file,
                                 parse_efield_file, parse_xrd_sf_file, parse_elf_fmt_file,
                                 parse_chdiff_fmt_file, parse_pot_fmt_file, parse_den_fmt_file,
                                 parse_elastic_file, parse_ts_file, parse_magres_file,
                                 parse_tddft_file, parse_err_file)

PARSERS = {
    "castep": parse_castep_file,
    "cell": parse_cell_param_file,
    "param": parse_cell_param_file,
    "geom": parse_md_geom_file,
    "md": parse_md_geom_file,
    "bands": parse_bands_file,
    "hug": parse_hug_file,
    "phonon_dos": parse_phonon_dos_file,
    "efield": parse_efield_file,
    "xrd_sf": parse_xrd_sf_file,
    "elf_fmt": parse_elf_fmt_file,
    "chdiff_fmt": parse_chdiff_fmt_file,
    "pot_fmt": parse_pot_fmt_file,
    "den_fmt": parse_den_fmt_file,
    "elastic": parse_elastic_file,
    "ts": parse_ts_file,
    "magres": parse_magres_file,
    "tddft": parse_tddft_file,
    "err": parse_err_file
    }
CASTEP_OUTPUT_NAMES = tuple(PARSERS.keys())
CASTEP_FILE_FORMATS = tuple(f".{typ}" for typ in CASTEP_OUTPUT_NAMES)



"""
Argument parser
"""
from typing import Sequence, Dict, List
from pathlib import Path
import argparse
# pylint: disable=line-too-long

from .utility import SUPPORTED_FORMATS
from .parsers import CASTEP_OUTPUT_NAMES, CASTEP_FILE_FORMATS

AP = argparse.ArgumentParser(
    prog="castep_outputs",
    description=f"""Attempts to find all files for seedname, filtered by `inc` args (default: all).
    Explicit files can be passed using longname arguments.
    castep_outputs can parse most human-readable castep outputs including: {', '.join(CASTEP_FILE_FORMATS)}"""
)

AP.add_argument("seedname", nargs=argparse.REMAINDER, help="Seed name for data")
AP.add_argument("-V", "--version", action="version", version="%(prog)s v0.1")
AP.add_argument("-L", "--log", help="Verbose output",
                choices=('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'), default="WARNING")
AP.add_argument("-o", "--output", help="File to write output, default: screen", default=None)
AP.add_argument("-f", "--out-format", help="Output format", choices=SUPPORTED_FORMATS, default="json")

AP.add_argument("-t", "--testing", action="store_true", help="Set testing mode to produce flat outputs")

AP.add_argument("-A", "--inc-all", action="store_true", help="Extract all available information")

for output_name in CASTEP_OUTPUT_NAMES:
    AP.add_argument(f"--inc-{output_name}", action="store_true", help=f"Extract .{output_name} information")

for output_name in CASTEP_OUTPUT_NAMES:
    AP.add_argument(f"--{output_name}", nargs="*",
                    help=f"Extract from {output_name.upper()} as .{output_name} type", default=[])


def parse_args(to_parse: Sequence[str] = ()) -> argparse.Namespace:
    """ Parse all arguments and add those caught by flags """
    args = AP.parse_args()

    parse_all = args.inc_all or not any(getattr(args, f"inc_{typ}") for typ in CASTEP_OUTPUT_NAMES)

    # Set all flags
    if parse_all and not to_parse:
        for typ in CASTEP_OUTPUT_NAMES:
            setattr(args, f"inc_{typ}", True)

    # Only parse those which are requested
    for typ in to_parse:
        setattr(args, f"inc_{typ}", True)

    # Add seeded files into parse list if to be included
    for seed in args.seedname:
        seed = Path(seed)
        if seed.is_file() and (ext := seed.suffix[1:]) in CASTEP_OUTPUT_NAMES:
            getattr(args, ext).append(str(seed))
        else:
            for typ in CASTEP_OUTPUT_NAMES:
                trial = seed.with_suffix(f".{typ}")
                if getattr(args, f"inc_{typ}") and trial.is_file():
                    getattr(args, typ).append(str(trial))

            if args.inc_err and (err_files := seed.parent.glob(f"{seed}.*.err")):
                args.err.extend(map(str, err_files))

    return args


def args_to_dict(args: argparse.Namespace) -> Dict[str, List[str]]:
    """ Convert args namespace to dictionary """
    out_dict = {typ: getattr(args, typ) for typ in CASTEP_OUTPUT_NAMES}
    return out_dict



"""
Parse castep .cell and .param files
"""
from typing import Dict, TextIO, List, Union, Tuple
from collections import defaultdict
import re

import castep_outputs.castep_res as REs
from .utility import (log_factory, to_type, determine_type, atreg_to_index)
from .castep_res import get_block
from .constants import AtomIndex, ThreeByThreeMatrix, ThreeVector


def parse_cell_param_file(cell_param_file: TextIO) -> List[Dict[str, Union[str, Dict[str, str]]]]:
    """ Parse .cell/.param files into dict ready to JSONise """

    logger = log_factory(cell_param_file)
    curr = {}

    for line in cell_param_file:
        # Strip comments
        line = re.split("[#!]", line)[0]

        if block := get_block(line, cell_param_file,
                              re.compile(r"^\s*%block ", re.IGNORECASE),
                              re.compile(r"^\s*%endblock", re.IGNORECASE)):

            block_title = next(block).split()[1].lower()

            logger("Found block %s", block_title)

            curr[block_title] = _PARSERS.get(block_title, _parse_general)(block)

        elif match := REs.PARAM_VALUE_RE.match(line):
            key, val, unit = match.group("key", "val", "unit")
            key = key.strip().lower()
            logger("Found param %s", key)

            if " " in val.strip():
                val = to_type(val.split(), determine_type(val))
            else:
                val = to_type(val, determine_type(val))
            if unit:
                unit = unit.strip()
                curr[key] = (val, unit)
            else:
                curr[key] = val

    return [curr]


parse_cell_file = parse_cell_param_file
parse_param_file = parse_cell_param_file


def _parse_devel_code_block(in_block: TextIO) -> Dict[str, Union[str, float, int]]:
    """ Parse devel_code block to dict """

    in_block = in_block.read()
    in_block = " ".join(in_block.splitlines())
    matches = re.finditer(REs.DEVEL_CODE_BLOCK_GENERIC_RE, in_block, re.IGNORECASE | re.MULTILINE)
    devel_code_parsed = {}

    # Get groups
    for blk in matches:
        block_title = blk.group(1)
        block_str = blk.group(0).split(":")[1]
        block = {}
        for par in re.finditer(REs.DEVEL_CODE_VAL_RE, block_str,
                               re.IGNORECASE | re.MULTILINE):

            key, val = re.split("[:=]", par.group(0))
            typ = determine_type(val)
            block[key] = to_type(val, typ)
            block_str = block_str.replace(par.group(0), "")

        if block_str.strip():
            block["data"] = []

        for par in block_str.split():
            block["data"].append(to_type(par, determine_type(par)))

        devel_code_parsed[block_title.lower()] = block

        # Remove matched to get remainder
        in_block = in_block.replace(blk.group(0), "")

    # Catch remainder
    for par in re.finditer(REs.DEVEL_CODE_VAL_RE, in_block):
        key, val = re.split("[:=]", par.group(0))
        typ = determine_type(val)
        devel_code_parsed[key] = to_type(val, typ)

    return devel_code_parsed


def _parse_ionic_constraints(block: TextIO) -> Dict[AtomIndex, Tuple[float, float, float]]:
    accum = defaultdict(list)

    for line in block:
        if match := REs.IONIC_CONSTRAINTS_RE.match(line):
            ind = atreg_to_index(match)

            accum[ind].append(to_type(match["val"].split(), float))

    return accum


def _parse_nonlinear_constraints(block: TextIO):
    accum = []

    for line in block:
        if re.match(r"^\s*%endblock", line, re.IGNORECASE):
            continue

        key, line = line.split(maxsplit=1)
        cons = {"key": key,
                "atoms": {atreg_to_index(match): to_type(match.group("x", "y", "z"), int)
                          for match in REs.ATDAT3VEC.finditer(line)}}
        accum.append(cons)

    return accum


def _parse_positions(block: TextIO) -> Dict[AtomIndex, Tuple[float, float, float]]:
    accum = {}
    cnt = defaultdict(lambda: 0)

    for line in block:
        if match := REs.POSITIONS_LINE_RE.match(line):
            spec = match["spec"]
            cnt[spec] += 1
            ind = (spec, cnt[spec])
            pos = to_type(match["pos"].split(), float)

            accum[ind] = {"pos": pos}

            if match := REs.POSITIONS_SPIN_RE.search(line):
                accum[ind]["spin"] = to_type(match["spin"], float)

            if match := REs.POSITIONS_MIXTURE_RE.search(line):
                accum[ind]["mixed"] = to_type(match["mix"], int)
                accum[ind]["ratio"] = to_type(match["ratio"], float)

    return accum


def _parse_hubbard_u(block: TextIO) -> Dict[Union[str, AtomIndex], Dict[str, float]]:
    accum = {}

    for line in block:
        if re.match(r"^\s*%endblock", line, re.IGNORECASE):
            continue

        if match := re.match(rf"^\s*({REs.ATOM_NAME_RE})\s*(\d+)?", line):
            spec = (match[1], int(match[2])) if match.lastindex == 2 else match[1]
            accum[spec] = {}

            line = line.replace(match[0], "")
            for key, val in re.findall(r"\b(?P<key>\S+)\s*:\s*(?P<val>\S+)\b", line):
                val = to_type(val, determine_type(val))
                accum[spec][key] = val
        else:
            accum["units"] = line.strip()

    return accum


def _parse_sedc(block: TextIO) -> Dict[str, Dict[str, float]]:
    accum = {}
    for line in block:
        if re.match(r"^\s*%endblock", line, re.IGNORECASE):
            continue

        spec, line = line.strip().split(maxsplit=1)
        accum[spec] = {}

        for key, val in re.findall(r"\b(?P<key>\S+)\s*:\s*(?P<val>\S+)\b", line):
            val = to_type(val, determine_type(val))
            accum[spec][key] = val

    return accum


def _parse_symops(block: TextIO) -> List[Dict[str, Union[ThreeByThreeMatrix, ThreeVector]]]:

    accum = []
    tmp = [to_type(numbers, float)
           for line in block
           if (numbers := REs.FLOAT_RAT_RE.findall(line))]

    accum = [{'r': tmp[i:i+3],
              't': tmp[i+3]}
             for i in range(0, len(tmp), 4)]

    return accum


def _parse_general(block: TextIO) -> Dict[str, Union[str, float]]:
    block_data = {"data": []}
    for line in block:
        line = re.split("[#!]", line)[0]

        if REs.SPEC_PROP_RE.match(line):
            if isinstance(block_data["data"], list):
                block_data["data"] = {}

            spec, val = line.strip().split(maxsplit=1)
            val = to_type(val, determine_type(val))

            block_data["data"][spec] = val

        elif numbers := REs.FLOAT_RAT_RE.findall(line):
            block_data["data"].append(to_type(numbers, float))

        elif re.match(r"^\s*%endblock", line, re.IGNORECASE):
            pass

        else:
            block_data["units"] = line.strip()

    return block_data


_PARSERS = {"devel_code": _parse_devel_code_block,
            "ionic_constraints": _parse_ionic_constraints,
            "nonlinear_constraints": _parse_nonlinear_constraints,
            "positions_abs": _parse_positions,
            "positions_frac": _parse_positions,
            "positions_abs_intermediate": _parse_positions,
            "positions_frac_intermediate": _parse_positions,
            "positions_abs_product": _parse_positions,
            "positions_frac_product": _parse_positions,
            "sedc_custom_params": _parse_sedc,
            "hubbard_u": _parse_hubbard_u,
            "symmetry_ops": _parse_symops}



"""
Run main castep parser
"""
from typing import Callable, Optional, TextIO, Union
from pathlib import Path
import logging
import io
import sys

from .args import (parse_args, args_to_dict)
from .utility import (normalise, json_safe, flatten_dict, get_dumpers)
from .constants import OutFormats
from .parsers import PARSERS


def parse_single(in_file: Union[Path, TextIO], parser: Callable = None,
                 out_format: OutFormats = "print",
                 *, loglevel: int = logging.WARNING, testing: bool = False):
    """
    Parse a file using the given parser and post-process according to options
    """

    logging.basicConfig(format="%(levelname)s: %(message)s", level=loglevel)

    if not isinstance(in_file, (io.TextIOBase, Path)):
        in_file = Path(in_file)

    if parser is None:
        ext = in_file.suffix.strip(".")

        if ext not in PARSERS:
            raise KeyError(f"Parser for file {in_file} (assumed type: {ext}) not found")

        parser = PARSERS[ext]

    if isinstance(in_file, io.TextIOBase):
        data = parser(in_file)
    else:
        with open(in_file, mode='r', encoding="utf-8") as file:
            data = parser(file)

    if out_format == "json" or testing:
        data = normalise(data, {dict: json_safe, complex: json_safe})
    elif out_format in ("yaml", "ruamel"):
        data = normalise(data, {tuple: list})
    else:
        data = normalise(data)

    if testing:
        if isinstance(data, list):
            data = [flatten_dict(run) for run in data]
        else:
            data = flatten_dict(data)

    return data


def parse_all(output: Optional[Path] = None, out_format: OutFormats = "json",
              *, loglevel: int = logging.WARNING, testing: bool = False, **files):
    """ Parse all files in files dict """
    file_dumper = get_dumpers(out_format)

    data = {}
    for typ, paths in files.items():
        parser = PARSERS[typ]
        for path in paths:
            data[path] = parse_single(path, parser, out_format, loglevel=loglevel, testing=testing)

    if len(data) == 1:
        data = data.popitem()[1]

    if output is None:
        file_dumper(data, sys.stdout)
        print()
    elif isinstance(output, io.TextIOBase):
        file_dumper(data, output)
    else:
        with open(output, 'a+', encoding='utf-8') as out_file:
            file_dumper(data, out_file)


def main():
    """ Run the main program from command line """
    args = parse_args()
    dict_args = args_to_dict(args)

    parse_all(output=args.output,
              loglevel=getattr(logging, args.log.upper()),
              testing=args.testing,
              out_format=args.out_format,
              **dict_args)


if __name__ == "__main__":
    main()
